import EventEmitter from "events";

import { ListenerRss } from "listener-rss";
import { SqliteTools } from "./sqlite-tools";

/**
 * Permit to manage a ListenerRSS array, data storage and event aggregation
 */
export class ListenerRssAggregator extends EventEmitter {
  private listenerMap: Map<string, ListenerRss>;
  private looprunning: boolean = false;

  private sqliteDb: SqliteTools;

  constructor(args: {
    sqliteDB: SqliteTools;
    configInstantiated: Map<string, ListenerRss>;
  }) {
    super();

    this.sqliteDb = args.sqliteDB;
    this.listenerMap = args.configInstantiated;
  }

  static async instantiateAggregator(
    path: string
  ): Promise<{
    sqliteDB: SqliteTools;
    configInstantiated: Map<string, ListenerRss>;
  }> {
    const sqliteDB = new SqliteTools(path);
    const configInstantiated = await this.loadDb(sqliteDB);
    return { sqliteDB, configInstantiated };
  }

  /**
   * @brief [private don't use it] create a sqlite instance for the specified path and load the data
   */
  private static async loadDb(
    sqliteDb: SqliteTools
  ): Promise<Map<string, ListenerRss>> {
    const config = await this.extractConfig(sqliteDb);

    const configInstantiated = new Map<string, ListenerRss>();

    for (const obj of config)
      configInstantiated.set(obj.address, new ListenerRss(obj));

    return configInstantiated;
  }

  /**
   * @brief [private don't use it] load the actual config inside the sqliteTools instance specified
   */
  private static async extractConfig(
    sqliteDb: SqliteTools
  ): Promise<ListenerRss.Config[]> {
    await sqliteDb.ensureTableExists();
    return await sqliteDb.fetchAll();
  }

  async getAllConfigs(): Promise<ListenerRss.Config[]> {
    return await ListenerRssAggregator.extractConfig(this.sqliteDb);
  }

  private addNewListener(info: ListenerRss.Config): ListenerRss {
    const newListener = new ListenerRss(info);
    this.listenerMap.set(newListener.address, newListener);

    newListener.on("update", (obj) =>
      this.emit("update", {
        ...obj,
        ...{ addressListener: newListener.address },
      })
    );
    newListener.on("newEntries", (obj) =>
      this.emit("newEntries", {
        ...obj,
        ...{ addressListener: newListener.address },
      })
    );
    newListener.on("error", (err) =>
      this.emit("error", {
        ...err,
        ...{ addressListener: newListener.address },
      })
    );

    return newListener;
  }

  private removeOldListener(adr: string): void {
    const oldListener = this.listenerMap.get(adr);

    if (!oldListener) return;

    oldListener.stop();
    oldListener
      .removeAllListeners("update")
      .removeAllListeners("newEntries")
      .removeAllListeners("error");
    this.listenerMap.delete(adr);
  }

  async registerListener(info: ListenerRss.Config) {
    if (this.listenerMap.has(info.address)) return;

    const listener = this.addNewListener(info);
    await this.sqliteDb.insertListener(listener);
  }

  async unregisterListener(adr: string): Promise<void> {
    if (!this.listenerMap.has(adr)) return;

    this.removeOldListener(adr);
    await this.sqliteDb.deleteListener(adr);
  }

  async saveOverride(expectedConfig: ListenerRss.Config[]) {
    const actualConfig = await this.getAllConfigs();

    for (const newItem of expectedConfig.filter(
      (item) => !actualConfig.includes(item)
    )) {
      await this.registerListener(newItem);
    }

    for (const oldItem of actualConfig.filter(
      (item) => !expectedConfig.includes(item)
    )) {
      await this.unregisterListener(oldItem.address);
    }
  }

  startAll(): void {
    if (this.looprunning) return;

    this.looprunning = true;
    for (const item of this.listenerMap.values()) item.start();
  }

  stopAll(): void {
    if (!this.looprunning) return;

    this.looprunning = false;
    for (const item of this.listenerMap.values()) item.stop();
  }
}
